--[[
Eightyfour, a World of Warcraft addon that announces when cc spells are broken
  and when misdirection/tricks of the trade are used.
]]--

local _G = getfenv(0)
local format = _G.string.format
local bit_band = _G.bit.band
local UnitClass = _G.UnitClass
local COMBATLOG_OBJECT_REACTION_FRIENDLY = _G.COMBATLOG_OBJECT_REACTION_FRIENDLY

local THROTTLE_TIME = 1

local Eightyfour = LibStub("AceAddon-3.0"):NewAddon("Eightyfour", "AceEvent-3.0")

local L = setmetatable(
GetLocale() == "deDE" and {
--[[ example.
	["Spells"] = "Spells",
	["Toggle announcements for individual spells."] = "Toggle announcements for individual spells.",
	["%s's %s"] = "%s's %s",
	["%s%s on %s%s%s|r removed by %s."] = "%s%s on %s%s%s|r removed by %s.",
	["%s cast %s on %s."] = "%s cast %s on %s.",
	["<unknown>"] = "<unknown>",
	["Announce to:"] = "Announce to:",
	["Channels:"] = "Channels:",
	["Say"] = "Say",
	["Yell"] = "Yell",
	["Raid"] = "Raid",
	["Party"] = "Party",
	["Guild"] = "Guild",
	["Officer"] = Officer",
	["Chat Frame"] = "Chat Frame",
	["Raid Warning"] = "Raid Warning",
	["Custom Channel"] = "Custom Channel",
]]--
} or GetLocale() == "ruRU" and {} or {}, {__index = function(t,k) return k end})

local classColors = {
	["Unknown"] = "|cffaaaaaa"
}
for k, v in pairs(RAID_CLASS_COLORS) do
	classColors[k] = "|cff" .. format("%02x%02x%02x", v.r*255, v.g*255, v.b*255)
end

local coloredName = setmetatable({}, {__index =
	function(t,k)
		if type(k) == "nil" then return L["<unknown>"] end
		local class = select(2, UnitClass(k)) or "Unknown"
		t[k] = classColors[class]  .. k .. "|r"
		return t[k]
	end
})

local spells = {
	118, -- Polymorph
	9484, --  Shackle Undead
	2637, -- Hibernate
	14309, -- Freezing Trap Effect
	60210, -- Freezing Arrow Effect
	6358, -- Seduction
	6770, -- Sap
	20066, -- Repentance
	51514, -- Hex
	34477, -- Misdirection
	57934, -- Tricks of the Trade
}

local icon
do
	local COMBATLOG_OBJECT_RAIDTARGET1 = _G.COMBATLOG_OBJECT_RAIDTARGET1
	local COMBATLOG_OBJECT_SPECIAL_MASK = _G.COMBATLOG_OBJECT_SPECIAL_MASK
	function icon(flags)
		if bit_band(flags, COMBATLOG_OBJECT_SPECIAL_MASK) ~= 0 then
			for i = 1,8 do
				local mask = COMBATLOG_OBJECT_RAIDTARGET1 * (2 ^ (i - 1))
				if bit_band(flags, mask) == mask then
					return Eightyfour.db.profile.output == "chat" and format("%s0\124t", _G.ICON_LIST[i]) or format("{rt%d}", i)
				end
			end
		end
		return ""
	end
end

local channels = {
	chat = L["Chat Frame"],
	say = L["Say"],
	yell = L["Yell"],
	raid = L["Raid"],
	party = L["Party"],
	guild = L["Guild"],
	officer = L["Officer"],
	raid_warning = L["Raid Warning"],
	channel = L["Custom Channel"],
}
local customChannels = {}
local inRaid = false

local defaults = {
	profile = {
		spells = {},
		output = "chat",
		channel = nil,
	}
}

local options = {
	type = 'group',
	name = 'Eightyfour',
	args = {
		announce = {
			type = 'select',
			name = L["Announce to:"],
			values = channels,
			get = function() return Eightyfour.db.profile.output end,
			set = function(_,v)
				Eightyfour.db.profile.output = v
				if v == "channel" then Eightyfour:AddChannels(GetChannelList()) end
			      end,
			order = 10,
		},
		channel = {
			type = 'select',
			disabled = function() return Eightyfour.db.profile.output ~= "channel" and true or false end,
			name = L["Channels:"],
			values = customChannels,
			get = function() return Eightyfour.db.profile.channel end,
			set = function(_,v) Eightyfour.db.profile.channel = v end,
			order = 11,
		},
		spells = {
			type = 'group',
			name = L["Spells"],
			desc = L["Toggle announcements for individual spells."],
			order = 20,
			args = {},
		},
	}
}

function Eightyfour:Announce(msg)
	if Eightyfour.db.profile.output == "chat" then
		print(msg)
		return
	end
	-- Need to sanitize all other messages
	msg = msg:gsub("(|c%x%x%x%x%x%x%x%x)", ""):gsub("(|r)", "")
	if Eightyfour.db.profile.output == "say"
	  or Eightyfour.db.profile.output == "yell"
	  or Eightyfour.db.profile.output == "raid"
	  or Eightyfour.db.profile.output == "party"
	  or Eightyfour.db.profile.output == "guild"
	  or Eightyfour.db.profile.output == "officer"
	  or Eightyfour.db.profile.output == "raid_warning" then
		SendChatMessage(msg, Eightyfour.db.profile.output)
	elseif Eightyfour.db.profile.output == "channel" then
		if not Eightyfour.db.profile.output then
			print("|cff33ff99Eightyfour:|r ERRROR: No channel specified to announce to.")
		else
			local channelName = GetChannelName(Eightyfour.db.profile.channel)
			if not channelName then
				print("|cff33ff99Eightyfour:|r ERRROR: Invalid channel id.")
			else
				SendChatMessage(msg, "CHANNEL", nil, channelName)
			end
		end		
	end
end

function Eightyfour:AddChannels(...)
	wipe(customChannels)
	for i = 1, select("#", ...), 2 do
		local id, name = select(i, ...)
		customChannels[tostring(id)] = name
	end
end

function Eightyfour:OnInitialize()
	for i, k in ipairs(spells) do
		local name, _, icon = GetSpellInfo(k)
		spells[name] = icon
		spells[i] = nil
		defaults.profile.spells[name] = true
		options.args.spells.args[name] = {
			type = 'toggle',
			name = name,
			get = function() return self.db.profile.spells[name] end,
			set = function(_,v) self.db.profile.spells[name] = v end,
		}
	end
	self.db = LibStub("AceDB-3.0"):New("EightyfourDB", defaults, "Default")
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Eightyfour", options)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Eightyfour", "Eightyfour")
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("RAID_ROSTER_UPDATE", function() inRaid = UnitInRaid("player") and true or false end)
	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE", function() Eightyfour:AddChannels(GetChannelList()) end)
	self:AddChannels(GetChannelList())
end


local throttle = 0
function Eightyfour:COMBAT_LOG_EVENT_UNFILTERED(_, _, eventType, _, srcName, _, _, dstName, dstFlags, spellID, spellName, _, _, extraName)
	if inRaid == false then return end
	if eventType == "SPELL_AURA_BROKEN_SPELL" or eventType == "SPELL_AURA_BROKEN" or eventType == "SPELL_DISPEL" then
		local spell = spells[spellName]
		if spell and self.db.profile.spells[spellName] and bit_band(dstFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY) == 0 then
			if (GetTime() - throttle) < THROTTLE_TIME then
				return
			end
			throttle = GetTime()
			local spellIcon = self.db.profile.output == "chat" and ("\124T%s:0\124t"):format(spell) or ""
			local brokenBy = extraName and format(L["%s's %s"], coloredName[srcName], extraName or L["<unknown>"]) or coloredName[srcName]
			self:Announce(format(L["%s%s on %s%s%s|r removed by %s."], spellIcon, spellName, icon(dstFlags), _G.RED_FONT_COLOR_CODE, dstName, brokenBy))
		end
	elseif eventType == "SPELL_CAST_SUCCESS" and (spellID == 34477 or spellID == 57934) then
		if self.db.profile.spells[spellName] then
			self:Announce(format(L["%s cast %s on %s."], coloredName[srcName], spellName, coloredName[dstName]))
		end
	end
end

function Eightyfour:OpenConfig() InterfaceOptionsFrame_OpenToCategory("Eightyfour") end
SlashCmdList["EIGHTYFOUR"] = function() Eightyfour:OpenConfig() end
_G["SLASH_EIGHTYFOUR1"] = "/eightyfour"
_G["SLASH_EIGHTYFOUR2"] = "/84"